(provide 'fibheap)

;; Fibonacci heap as described in CLRS 2nd Edition, p. 476
;;
;; This code is placed in the public domain by Mark Wutka
;;

;; The minimal usage is:

;; Create a new heap
;(defparameter myheap (make-fib-heap))

;; Add some data
;let ((inserted-node (fib-heap-insert myheap {a data node here})))

;; Extract the next item from the heap
;(let ((next (fib-heap-extract-min myheap)))

;; The fib-heap-extract-min function returns nil when there
;; is no more data. You can also examine (fib-heap-size myheap)
;; to see if there is more data available.

;; You can also adjust the value of a key downward, but not upward.
;; If you change the key value of a node, call:
;; (fib-heap-decrease-key myheap mynode)
;; Make sure you have already adjusted the key value. Since you
;; have to call this on the node itself and not on the data item,
;; you may need to keep track of the nodes. The fin-heap-insert function
;; returns the heap node after it creates it.

;; The Fibonacci heap is particularly suited for Dijkstra's shortest path
;; first algorithm. You can insert values very quickly ( O(1) time).
;; Extracting values is O(log n).
;; 

;; A fib-heap-node represents a single node in the heap
;; Nodes are stored in a circular linked-list and may point
;; to a child node
(defstruct (fib-heap-node (:print-function print-fib-heap-node))
  (data nil)
  (parent nil)
  (left nil)
  (right nil)
  (child nil)
  (degree 0)
  (mark nil))

;; Need a special print function. The default print function goes
;; berzerk with the circular list
(defun print-fib-heap-node (n stream depth)
  (when (< depth 20)
    (if (null n)
	(print "NIL")
	(progn
	  (dotimes (d depth) (format stream " "))
	  (format stream "fib-heap-node(~a)~%" (fib-heap-node-data n))
	  (map-fib-heap-node-list #'(lambda (x) (print-fib-heap-node x stream (+ depth 1)))
				  (fib-heap-node-child n))))))

;; The heap itself keeps track of its size, the minimum node (i.e. the root of the tree)
;; and a function to compare nodes.
(defstruct (fib-heap (:print-function print-fib-heap))
  (size 0)
  (min nil)
  (compare-func #'<))

;; Again override the print functio nto handle circular lists
(defun print-fib-heap (h stream depth)
  (format stream "fib-heap size(~a)~%" (fib-heap-size h))
  (map-fib-heap-node-list #'(lambda (x) (print-fib-heap-node x stream (+ depth 1)))
			  (fib-heap-min h)))


(defun map-fib-heap-node-list (func list)
  "Execute (func x) for each x in the circular list"
  (when list
    (let ((start list))
      (loop do
	   (funcall func list)
	   (setf list (fib-heap-node-left list))
	   until (eql list start)))))

(defun fib-heap-node-list-length (list)
  "Compute the length of the circular list"
  (let ((len 0)
	(start list))
    (when start
      (loop do
	   (incf len)
	   (setf list (fib-heap-node-left list))
	   until (eql list start)))
    len))

(defun create-fib-heap-node (data)
  "Create a new heap node and initialize the left and right pointers"
  (let ((new-node (make-fib-heap-node)))
    (setf (fib-heap-node-data new-node) data)
    (setf (fib-heap-node-left new-node) new-node)
    (setf (fib-heap-node-right new-node) new-node)
    new-node))

(defun fib-heap-splice-lists (list1 list2)
  "Splice two circular lists together"
  (setf (fib-heap-node-left (fib-heap-node-right list1)) list2)
  (setf (fib-heap-node-left (fib-heap-node-right list2)) list1)
  (let ((temp (fib-heap-node-right list1)))
    (setf (fib-heap-node-right list1) (fib-heap-node-right list2))
    (setf (fib-heap-node-right list2) temp)))

(defun fib-heap-insert-list-node (list node)
  "Insert a new node into a circular list"
  (setf (fib-heap-node-left node) node)
  (setf (fib-heap-node-right node) node)
  (fib-heap-splice-lists list node))

(defun fib-heap-node-add-child (node new-child)
  "Add a new child to a node"
  (cond ((null (fib-heap-node-child node))
	 (setf (fib-heap-node-child node) new-child))
	(t (fib-heap-splice-lists (fib-heap-node-child node) new-child)))
  (setf (fib-heap-node-parent new-child) node)
  (incf (fib-heap-node-degree node)))

(defun fib-heap-remove-list-node (node)
  "Remove the specified node from whatever list it belongs to."
  (cond ((null node) nil)
	((eql node (fib-heap-node-left node)) nil)
	(t
	 (setf (fib-heap-node-right (fib-heap-node-left node))
	       (fib-heap-node-right node))
	 (setf (fib-heap-node-left (fib-heap-node-right node))
	       (fib-heap-node-left node))
	 (setf (fib-heap-node-left node) node)
	 (setf (fib-heap-node-right node) node)
	 t)))
	

(defun fib-heap-insert (heap data)
  "Insert data into the heap"
  (incf (fib-heap-size heap))
  (cond ((fib-heap-min heap)
	 (let ((new-node (create-fib-heap-node data)))
	   (fib-heap-splice-lists (fib-heap-min heap) new-node)
	   (when (funcall (fib-heap-compare-func heap) data 
			  (fib-heap-node-data (fib-heap-min heap)))
	     (setf (fib-heap-min heap) new-node))
	   new-node))
	(t (setf (fib-heap-min heap) (create-fib-heap-node data)))))


(defun fib-heap-union (heap1 heap2)
  (cond ((and heap1 heap2)
	 (fib-heap-splice-lists (fib-heap-min heap1) (fib-heap-min heap2))
	 (when (funcall (fib-heap-compare-func heap1)
			(fib-heap-node-data (fib-heap-min heap2)) 
			(fib-heap-node-data (fib-heap-min heap1)))
	   (setf (fib-heap-min heap1) (fib-heap-min heap2)))
	 (setf (fib-heap-size heap1)
	       (+ (fib-heap-size heap1) (fib-heap-size heap2)))
	 heap1)
	(heap2 heap2)
	(t heap1)))

(defun fib-heap-extract-min (heap)
  "Remove the minimum value from the heap and return it"
  (cond ((fib-heap-min heap)
	 (let ((z (fib-heap-min heap))
	       (z-right nil))
	   (map-fib-heap-node-list #'(lambda (n) (setf (fib-heap-node-parent n) nil))
				   (fib-heap-node-child z))
	   (when (fib-heap-node-child z)
	     (fib-heap-splice-lists (fib-heap-min heap) (fib-heap-node-child z)))
	   
	   (setf z-right (fib-heap-node-right z))
	   (if (fib-heap-remove-list-node z)
	       (setf (fib-heap-min heap) z-right)
	       (setf (fib-heap-min heap) (fib-heap-node-child z)))
	   (when (fib-heap-min heap)
	     (fib-heap-consolidate heap))
	   (decf (fib-heap-size heap))
	   (fib-heap-node-data z)))
	(t nil)))

(defun fib-heap-link (x y)
  "Make y a child of x in the heap"
  (fib-heap-remove-list-node y)
  (fib-heap-node-add-child x y)
  (setf (fib-heap-node-mark y) nil))

(defun fib-heap-consolidate-2 (heap node)
  "Reconstruct the root level of the heap"
  (when node
    (setf (fib-heap-node-left node) node)
    (setf (fib-heap-node-right node) node)

    (cond ((null (fib-heap-min heap))
	   (setf (fib-heap-min heap) node))
	  ((funcall (fib-heap-compare-func heap)
		    (fib-heap-node-data node) (fib-heap-node-data (fib-heap-min heap)))
	   (fib-heap-splice-lists (fib-heap-min heap) node)
	   (setf (fib-heap-min heap) node))
	  (t (fib-heap-splice-lists (fib-heap-min heap) node)))))

(defun fib-heap-consolidate (heap)
  "Consolidate the heap so that each root node has a different number of children"
  (let* ((dh (ceiling (log (fib-heap-size heap) 2)))
	 (a (make-array (+ dh 1) :initial-element nil)))
    (do* ((num-roots (fib-heap-node-list-length (fib-heap-min heap)) (decf num-roots))
	  (x (fib-heap-min heap) next)
	  (d (fib-heap-node-degree x) (fib-heap-node-degree x))
	  (next (fib-heap-node-right x) (fib-heap-node-right x)))
	 ((<= num-roots 0))
      (do ((y (aref a d) (aref a d)))
	  ((null y))
	(when (funcall (fib-heap-compare-func heap) 
		       (fib-heap-node-data y) (fib-heap-node-data x))
	  (let ((temp x))
	    (setf x y)
	    (setf y temp)))
	(fib-heap-link x y)
	(setf (aref a d) nil)
	(incf d))
      (setf (aref a d) x))
    (setf (fib-heap-min heap) nil)
    (map nil #'(lambda (n) (fib-heap-consolidate-2 heap n)) a)))

(defun fib-heap-decrease-key (heap x)
  "Adjust the heap to account for the decrease in the key value of x"
  (let ((y (fib-heap-node-parent x)))
    (when (and y (funcall (fib-heap-compare-func heap) 
			  (fib-heap-node-data x) (fib-heap-node-data y)))
      (fib-heap-cut heap x y)
      (fib-heap-cascading-cut heap y)
    (when (funcall (fib-heap-compare-func heap)
		   (fib-heap-node-data x) (fib-heap-node-data (fib-heap-min heap)))
      (setf (fib-heap-min heap) x)))))

(defun fib-heap-cut (heap x y)
  "Move x out from being a child of y and move it to the root of the heap"
  (when (eql x (fib-heap-node-child y))
    (setf (fib-heap-node-child y) (fib-heap-node-right x)))
  (fib-heap-remove-list-node x)
  (decf (fib-heap-node-degree y))
  (when (= (fib-heap-node-degree y) 0)
    (setf (fib-heap-node-child y) nil))
  (fib-heap-insert-list-node (fib-heap-min heap) x)
  (setf (fib-heap-node-parent x) nil)
  (setf (fib-heap-node-mark x) nil))

(defun fib-heap-cascading-cut (heap y)
  "Cut any additional nodes that need to move to the root of the heap"
  (let ((z (fib-heap-node-parent y)))
    (when z
      (if (not (fib-heap-node-mark y))
	  (setf (fib-heap-node-mark y) t)
	  (progn
	    (fib-heap-cut heap y z)
	    (fib-heap-cascading-cut heap z))))))